from dataclasses import asdict

from itkwasm import PolyData


def test_polydata():
    itkwasm_polydata = PolyData()
    assert isinstance(itkwasm_polydata.metadata, dict)
    itkwasm_polydata_dict = asdict(itkwasm_polydata)



def test_itkwasm_js_package_config():
    from itkwasm.pyodide import JsPackageConfig

    module_url = "https://cdn.jsdelivr.net/npm/@itk-wasm/compress-stringify@0.4.2/dist/bundles/compress-stringify.js"
    pipelines_base_url = "https://cdn.jsdelivr.net/npm/@itk-wasm/compress-stringify@0.4.2/dist/pipelines"
    pipeline_worker_url = (
        "https://cdn.jsdelivr.net/npm/@itk-wasm/compress-stringify@0.4.2/dist/web-workers/pipeline.worker.js"
    )

    config = JsPackageConfig(module_url, pipelines_base_url, pipeline_worker_url)
    assert config.module_url == module_url
    assert config.pipelines_base_url == pipelines_base_url
    assert config.pipeline_worker_url == pipeline_worker_url

    config = JsPackageConfig(module_url)
    assert config.module_url == module_url
    assert config.pipelines_base_url is None
    assert config.pipeline_worker_url is None



import numpy as np

import pytest

from itkwasm import array_like_to_numpy_array

pytest.importorskip("cupy")


def test_array_like_to_numpy_array():
    arr = np.array([1, 2, 3])
    result = array_like_to_numpy_array(arr)
    assert isinstance(result, np.ndarray)
    assert np.array_equal(result, arr)

    import cupy as cp

    arr = cp.array([1, 2, 3])
    result = array_like_to_numpy_array(arr)
    assert isinstance(result, np.ndarray)
    assert np.array_equal(result, arr.get())



from enum import Enum


class IntTypes(str, Enum):
    Int8 = "int8"
    UInt8 = "uint8"
    Int16 = "int16"
    UInt16 = "uint16"
    Int32 = "int32"
    UInt32 = "uint32"
    Int64 = "int64"
    UInt64 = "uint64"

    SizeValueType = "uint64"
    IdentifierType = "uint64"
    IndexValueType = "int64"
    OffsetValueType = "int64"

    def __str__(self):
        return str(self.value)

    def __repr__(self):
        return f'"{str(self.value)}"'



from pathlib import Path, PurePosixPath
import tempfile
from dataclasses import asdict
import sys

import pytest

import itk
import numpy as np

from itkwasm import (
    InterfaceTypes,
    TextStream,
    BinaryStream,
    PipelineInput,
    PipelineOutput,
    Pipeline,
    TextFile,
    BinaryFile,
    Image,
    Mesh,
    Transform,
    TransformList,
    TransformParameterizations,
    FloatTypes,
)

test_input_dir = Path(__file__).resolve().parent / "input"
test_baseline_dir = Path(__file__).resolve().parent / "baseline"


def test_stdout_stderr():
    pipeline = Pipeline(test_input_dir / "stdout-stderr-test.wasi.wasm")
    pipeline.run([])

    # Test re-run
    pipeline.run([])


def test_pipeline_bytes():
    pipeline_path = test_input_dir / "stdout-stderr-test.wasi.wasm"
    with open(pipeline_path, "rb") as fp:
        wasm_bytes = fp.read()
    pipeline = Pipeline(wasm_bytes)
    pipeline.run([])


def test_pipeline_input_output_streams():
    pipeline = Pipeline(test_input_dir / "input-output-files-test.wasi.wasm")

    pipeline_inputs = [
        PipelineInput(InterfaceTypes.TextStream, TextStream("The answer is 42.")),
        PipelineInput(InterfaceTypes.BinaryStream, BinaryStream(bytes([222, 173, 190, 239]))),
    ]

    pipeline_outputs = [
        PipelineOutput(InterfaceTypes.TextStream),
        PipelineOutput(InterfaceTypes.BinaryStream),
    ]

    args = [
        "--memory-io",
        "--input-text-stream",
        "0",
        "--input-binary-stream",
        "1",
        "--output-text-stream",
        "0",
        "--output-binary-stream",
        "1",
    ]

    outputs = pipeline.run(args, pipeline_outputs, pipeline_inputs)

    # Test re-run
    outputs = pipeline.run(args, pipeline_outputs, pipeline_inputs)

    assert outputs[0].type == InterfaceTypes.TextStream
    assert outputs[0].data.data == "The answer is 42."
    assert outputs[1].type, InterfaceTypes.BinaryStream
    assert outputs[1].data.data[0], 222
    assert outputs[1].data.data[1], 173
    assert outputs[1].data.data[2], 190
    assert outputs[1].data.data[3], 239


@pytest.mark.skipif(
    sys.platform == "win32",
    reason="Windows tempfile resource, https://github.com/bytecodealliance/wasmtime-py/issues/132",
)
def test_pipeline_input_output_files():
    pipeline = Pipeline(test_input_dir / "input-output-files-test.wasi.wasm")
    input_text_file = PurePosixPath(test_input_dir / "input.txt")
    input_binary_file = PurePosixPath(test_input_dir / "input.bin")

    with tempfile.TemporaryDirectory() as tmpdirname:
        output_text_file = PurePosixPath(Path(tmpdirname) / "output.txt")
        output_binary_file = PurePosixPath(Path(tmpdirname) / "output.bin")

        pipeline_inputs = [
            PipelineInput(InterfaceTypes.TextFile, TextFile(input_text_file)),
            PipelineInput(InterfaceTypes.BinaryFile, BinaryFile(input_binary_file)),
        ]

        pipeline_outputs = [
            PipelineOutput(InterfaceTypes.TextFile, TextFile(output_text_file)),
            PipelineOutput(InterfaceTypes.BinaryFile, BinaryFile(output_binary_file)),
        ]

        args = [
            "--memory-io",
            "--use-files",
            "--input-text-file",
            str(input_text_file),
            "--input-binary-file",
            str(input_binary_file),
            "--output-text-file",
            str(output_text_file),
            "--output-binary-file",
            str(output_binary_file),
        ]

        outputs = pipeline.run(args, pipeline_outputs, pipeline_inputs)

        # Test re-run
        outputs = pipeline.run(args, pipeline_outputs, pipeline_inputs)

        assert outputs[0].type == InterfaceTypes.TextFile
        with open(outputs[0].data.path, "r") as fp:
            content = fp.read()
            assert content == "The answer is 42."
        assert outputs[1].type, InterfaceTypes.BinaryFile
        with open(outputs[1].data.path, "rb") as fp:
            content = fp.read()
            assert content[0] == 222
            assert content[1] == 173
            assert content[2] == 190
            assert content[3] == 239


def test_pipeline_write_read_image():
    pipeline = Pipeline(test_input_dir / "median-filter-test.wasi.wasm")

    data = test_input_dir / "cthead1.png"
    itk_image = itk.imread(data, itk.UC)
    itk_image_dict = itk.dict_from_image(itk_image)
    itkwasm_image = Image(**itk_image_dict)

    pipeline_inputs = [
        PipelineInput(InterfaceTypes.Image, itkwasm_image),
    ]

    pipeline_outputs = [
        PipelineOutput(InterfaceTypes.Image),
    ]

    args = [
        "--memory-io",
        "0",
        "0",
        "--radius",
        "2",
    ]

    outputs = pipeline.run(args, pipeline_outputs, pipeline_inputs)

    out_image = itk.image_from_dict(asdict(outputs[0].data))
    # To be addressed in itk-5.3.1
    out_image.SetRegions([256, 256])

    baseline = itk.imread(test_baseline_dir / "test_pipeline_write_read_image.png")

    difference = np.sum(itk.comparison_image_filter(out_image, baseline))
    assert difference == 0.0


def test_pipeline_dask_array_input():
    pipeline = Pipeline(test_input_dir / "median-filter-test.wasi.wasm")

    data = test_input_dir / "cthead1.png"
    itk_image = itk.imread(data, itk.UC)
    itk_image_dict = itk.dict_from_image(itk_image)
    itkwasm_image = Image(**itk_image_dict)

    from dask.array import from_array

    itkwasm_image.data = from_array(itkwasm_image.data)

    pipeline_inputs = [
        PipelineInput(InterfaceTypes.Image, itkwasm_image),
    ]

    pipeline_outputs = [
        PipelineOutput(InterfaceTypes.Image),
    ]

    args = [
        "--memory-io",
        "0",
        "0",
        "--radius",
        "2",
    ]

    outputs = pipeline.run(args, pipeline_outputs, pipeline_inputs)

    out_image = itk.image_from_dict(asdict(outputs[0].data))
    # To be addressed in itk-5.3.1
    out_image.SetRegions([256, 256])

    baseline = itk.imread(test_baseline_dir / "test_pipeline_write_read_image.png")

    difference = np.sum(itk.comparison_image_filter(out_image, baseline))
    assert difference == 0.0


def test_pipeline_write_read_mesh():
    pipeline = Pipeline(test_input_dir / "mesh-read-write-test.wasi.wasm")

    data = test_input_dir / "cow.vtk"
    itk_mesh = itk.meshread(data)
    itk_mesh_dict = itk.dict_from_mesh(itk_mesh)
    itkwasm_mesh = Mesh(**itk_mesh_dict)

    pipeline_inputs = [
        PipelineInput(InterfaceTypes.Mesh, itkwasm_mesh),
    ]

    pipeline_outputs = [
        PipelineOutput(InterfaceTypes.Mesh),
    ]

    args = [
        "--memory-io",
        "0",
        "0",
    ]

    outputs = pipeline.run(args, pipeline_outputs, pipeline_inputs)

    out_mesh_dict = asdict(outputs[0].data)
    # Native ITK Python binaries require uint64
    out_mesh_dict["cells"] = out_mesh_dict["cells"].astype(np.uint64)
    out_mesh_dict["meshType"]["cellComponentType"] = "uint64"
    out_mesh = itk.mesh_from_dict(out_mesh_dict)

    assert out_mesh.GetNumberOfPoints() == 2903
    assert out_mesh.GetNumberOfCells() == 3263


def test_pipeline_write_read_polydata():
    pipeline = Pipeline(test_input_dir / "mesh-to-poly-data.wasi.wasm")

    data = test_input_dir / "cow.vtk"
    itk_mesh = itk.meshread(data)
    itk_mesh_dict = itk.dict_from_mesh(itk_mesh)
    itkwasm_mesh = Mesh(**itk_mesh_dict)

    pipeline_inputs = [
        PipelineInput(InterfaceTypes.Mesh, itkwasm_mesh),
    ]

    pipeline_outputs = [
        PipelineOutput(InterfaceTypes.PolyData),
    ]

    args = [
        "--memory-io",
        "0",
        "0",
    ]

    outputs = pipeline.run(args, pipeline_outputs, pipeline_inputs)
    polydata = outputs[0].data

    pipeline = Pipeline(test_input_dir / "poly-data-to-mesh.wasi.wasm")

    pipeline_inputs = [
        PipelineInput(InterfaceTypes.PolyData, polydata),
    ]

    pipeline_outputs = [
        PipelineOutput(InterfaceTypes.Mesh),
    ]

    args = [
        "--memory-io",
        "0",
        "0",
    ]

    outputs = pipeline.run(args, pipeline_outputs, pipeline_inputs)

    out_mesh_dict = asdict(outputs[0].data)

    # native itk python binaries require uint64
    out_mesh_dict["cells"] = out_mesh_dict["cells"].astype(np.uint64)
    out_mesh_dict["meshType"]["cellComponentType"] = "uint64"
    assert np.isclose(out_mesh_dict["points"][0], 3.71636)
    out_mesh = itk.mesh_from_dict(out_mesh_dict)

    assert out_mesh.GetNumberOfPoints() == 2903
    assert out_mesh.GetNumberOfCells() == 3263

def test_pipeline_write_read_transform():
    pipeline = Pipeline(test_input_dir / "transform-read-write-test.wasi.wasm")

    data = test_input_dir / "LinearTransform.h5"
    itk_transform = itk.transformread(data)
    itk_transform_list = itk.dict_from_transform(itk_transform)
    itkwasm_transform_list = [Transform(**t) for t in itk_transform_list]

    pipeline_inputs = [
        PipelineInput(InterfaceTypes.TransformList, itkwasm_transform_list),
    ]

    pipeline_outputs = [
        PipelineOutput(InterfaceTypes.TransformList),
    ]

    args = [
        "--memory-io",
        "0",
        "0",
    ]

    outputs = pipeline.run(args, pipeline_outputs, pipeline_inputs)
    transform_list = outputs[0].data

    assert len(transform_list) == 1
    transform = transform_list[0]
    assert transform.transformType.transformParameterization == TransformParameterizations.Affine
    assert transform.transformType.parametersValueType == FloatTypes.Float64
    assert transform.numberOfParameters == 12
    assert transform.numberOfFixedParameters == 3
    np.testing.assert_allclose(transform.fixedParameters, np.array([0.0, 0.0, 0.0]))
    np.testing.assert_allclose(transform.parameters, np.array([
      0.65631490118447, 0.5806583745824385, -0.4817536741017158,
      -0.7407986817430222, 0.37486398378429736, -0.5573995934598175,
      -0.14306664045479867, 0.7227121458012518, 0.676179776908723,
      -65.99999999999997, 69.00000000000004, 32.000000000000036]))


from pathlib import Path

import itk

from itkwasm import Image
from itkwasm import FloatTypes, PixelTypes
import numpy as np

from itkwasm import cast_image


def test_cast_image_component():
    data = Path(__file__).absolute().parent / "input" / "cthead1.png"
    itk_image = itk.imread(data, itk.UC)
    itk_image_dict = itk.dict_from_image(itk_image)
    itkwasm_image = Image(**itk_image_dict)

    itkwasm_image_double = cast_image(itkwasm_image, component_type=FloatTypes.Float64)
    assert itkwasm_image_double.imageType.componentType == FloatTypes.Float64
    assert np.array_equal(itkwasm_image.data, itkwasm_image_double.data)


def test_cast_image_pixel_type():
    data = Path(__file__).absolute().parent / "input" / "cthead1.png"
    itk_image = itk.imread(data, itk.UC)
    itk_image_dict = itk.dict_from_image(itk_image)
    itkwasm_image = Image(**itk_image_dict)

    itkwasm_image_vector = cast_image(itkwasm_image, pixel_type=PixelTypes.VariableLengthVector)
    assert itkwasm_image_vector.imageType.pixelType == PixelTypes.VariableLengthVector
    assert np.array_equal(itkwasm_image.data, itkwasm_image_vector.data)



from dataclasses import dataclass, asdict
from typing import Optional

from .image import Image, ImageType
from .point_set import PointSet, PointSetType
from .mesh import Mesh, MeshType
from .polydata import PolyData, PolyDataType
from .transform import Transform, TransformType
from .binary_file import BinaryFile
from .binary_stream import BinaryStream
from .text_file import TextFile
from .text_stream import TextStream
from .float_types import FloatTypes
from .int_types import IntTypes
from .json_compatible import JsonCompatible
from .to_numpy_array import buffer_to_numpy_array


@dataclass
class JsPackageConfig:
    module_url: str
    pipelines_base_url: Optional[str] = None
    pipeline_worker_url: Optional[str] = None


class JsPackage:
    def __init__(self, config: JsPackageConfig):
        self._config = config
        self._js_module = None

    @property
    def config(self):
        return self._config

    @config.setter
    def config(self, value):
        self._config = value

    @property
    async def js_module(self):
        if self._js_module is not None:
            return self._js_module
        from pyodide.code import run_js

        js_module = await run_js(f"import('{self._config.module_url}')")
        if self._config.pipelines_base_url is not None:
            js_module.setPipelinesBaseUrl(self._config.pipelines_base_url)
        if self._config.pipeline_worker_url is not None:
            js_module.setPipelineWorkerUrl(self._config.pipeline_worker_url)
        self._js_module = js_module
        return js_module


class JsResources:
    def __init__(self):
        self._web_worker = None

    @property
    def web_worker(self):
        return self._web_worker

    @web_worker.setter
    def web_worker(self, value):
        self._web_worker = value


js_resources = JsResources()


def to_py(js_proxy):
    import pyodide

    if hasattr(js_proxy, "constructor") and js_proxy.constructor.name == "Uint8Array":
        return js_proxy.to_bytes()
    elif isinstance(js_proxy, pyodide.ffi.JsArray):
        return [to_py(value) for value in js_proxy]
    elif hasattr(js_proxy, "imageType"):
        image_dict = js_proxy.to_py()
        image_type = ImageType(**image_dict["imageType"])
        image_dict["imageType"] = image_type
        dimension = image_type.dimension
        component_type = image_type.componentType
        image_dict["direction"] = buffer_to_numpy_array(str(FloatTypes.Float64), image_dict["direction"]).reshape(
            (dimension, dimension)
        )
        shape = list(image_dict["size"])[::-1]
        if image_type.components > 1:
            shape.append(image_type.components)
        if image_dict["data"] is not None:
            image_dict["data"] = buffer_to_numpy_array(component_type, image_dict["data"]).reshape(tuple(shape))
        return Image(**image_dict)
    elif hasattr(js_proxy, "pointSetType"):
        point_set_dict = js_proxy.to_py()
        point_set_type = PointSetType(**point_set_dict["pointSetType"])
        point_set_dict["pointSetType"] = point_set_type
        dimension = point_set_type.dimension
        point_component_type = point_set_type.pointComponentType
        point_pixel_component_type = point_set_type.pointPixelComponentType
        if point_set_dict["points"] is not None:
            point_set_dict["points"] = buffer_to_numpy_array(point_component_type, point_set_dict["points"]).reshape(
                (-1, dimension)
            )
        if point_set_dict["pointData"] is not None:
            point_set_dict["pointData"] = buffer_to_numpy_array(point_pixel_component_type, point_set_dict["pointData"])
        return PointSet(**point_set_dict)
    elif hasattr(js_proxy, "meshType"):
        mesh_dict = js_proxy.to_py()
        mesh_type = MeshType(**mesh_dict["meshType"])
        mesh_dict["meshType"] = mesh_type
        dimension = mesh_type.dimension
        point_component_type = mesh_type.pointComponentType
        point_pixel_component_type = mesh_type.pointPixelComponentType
        cell_component_type = mesh_type.cellComponentType
        cell_pixel_component_type = mesh_type.cellPixelComponentType
        if mesh_dict["points"] is not None:
            mesh_dict["points"] = buffer_to_numpy_array(point_component_type, mesh_dict["points"]).reshape(
                (-1, dimension)
            )
        if mesh_dict["pointData"] is not None:
            mesh_dict["pointData"] = buffer_to_numpy_array(point_pixel_component_type, mesh_dict["pointData"])
        if mesh_dict["cells"] is not None:
            mesh_dict["cells"] = buffer_to_numpy_array(cell_component_type, mesh_dict["cells"])
        if mesh_dict["cellData"] is not None:
            mesh_dict["cellData"] = buffer_to_numpy_array(cell_pixel_component_type, mesh_dict["cellData"])
        return Mesh(**mesh_dict)
    elif hasattr(js_proxy, "polyDataType"):
        polydata_dict = js_proxy.to_py()
        polydata_type = PolyDataType(**polydata_dict["polyDataType"])
        polydata_dict["polyDataType"] = polydata_type
        point_pixel_component_type = polydata_type.pointPixelComponentType
        cell_pixel_component_type = polydata_type.cellPixelComponentType
        if polydata_dict["points"] is not None:
            polydata_dict["points"] = buffer_to_numpy_array(str(FloatTypes.Float32), polydata_dict["points"]).reshape(
                (-1, 3)
            )
        if polydata_dict["vertices"] is not None:
            polydata_dict["vertices"] = buffer_to_numpy_array(str(IntTypes.UInt32), polydata_dict["vertices"])
        if polydata_dict["lines"] is not None:
            polydata_dict["lines"] = buffer_to_numpy_array(str(IntTypes.UInt32), polydata_dict["lines"])
        if polydata_dict["polygons"] is not None:
            polydata_dict["polygons"] = buffer_to_numpy_array(str(IntTypes.UInt32), polydata_dict["polygons"])
        if polydata_dict["triangleStrips"] is not None:
            polydata_dict["triangleStrips"] = buffer_to_numpy_array(
                str(IntTypes.UInt32), polydata_dict["triangleStrips"]
            )
        if polydata_dict["pointData"] is not None:
            polydata_dict["pointData"] = buffer_to_numpy_array(point_pixel_component_type, polydata_dict["pointData"])
        if polydata_dict["cellData"] is not None:
            polydata_dict["cellData"] = buffer_to_numpy_array(cell_pixel_component_type, polydata_dict["cellData"])
        return PolyData(**polydata_dict)
    elif hasattr(js_proxy, "transformType"):
        transform_dict = js_proxy.to_py()
        transform_type = TransformType(**transform_dict["transformType"])
        transform_dict["transformType"] = transform_type
        parameters_value_type = transform_type.parametersValueType
        if transform_dict["fixedParameters"] is not None:
            transform_dict["fixedParameters"] = buffer_to_numpy_array(
                parameters_value_type, transform_dict["fixedParameters"]
            )
        if transform_dict["parameters"] is not None:
            transform_dict["parameters"] = buffer_to_numpy_array(
                parameters_value_type, transform_dict["parameters"]
            )
        return Transform(**transform_dict)
    elif hasattr(js_proxy, "path") and hasattr(js_proxy, "data") and isinstance(js_proxy.data, str):
        with open(js_proxy.path, "w") as fp:
            fp.write(js_proxy.data)
        return TextFile(path=js_proxy.path)
    elif hasattr(js_proxy, "path") and hasattr(js_proxy, "data"):
        with open(js_proxy.path, "wb") as fp:
            js_proxy.data.to_file(fp)
        return BinaryFile(path=js_proxy.path)
    elif hasattr(js_proxy, "data") and isinstance(js_proxy.data, str):
        text_stream_dict = js_proxy.to_py()
        return TextStream(**text_stream_dict)
    elif hasattr(js_proxy, "data"):
        binary_stream_dict = js_proxy.to_py()
        binary_stream_dict["data"] = bytes(binary_stream_dict["data"])
        return BinaryStream(**binary_stream_dict)
    elif isinstance(js_proxy, pyodide.ffi.JsProxy):
        return js_proxy.to_py()
    # int, etc
    return js_proxy


def to_js(py, **kwargs):
    import pyodide
    import js

    if isinstance(py, list):
        js_array = pyodide.ffi.to_js([])
        for value in py:
            js_array.append(to_js(value))
        return js_array
    elif isinstance(py, Image):
        image_dict = asdict(py)
        image_dict["direction"] = image_dict["direction"].ravel()
        if image_dict["data"] is not None:
            image_dict["data"] = image_dict["data"].ravel()
        if "metadata" in image_dict:
            image_dict["metadata"] = pyodide.ffi.to_js(image_dict["metadata"], dict_converter=js.Map.new)
        else:
            image_dict["metadata"] = pyodide.ffi.to_js(dict(), dict_converter=js.Map.new)
        return pyodide.ffi.to_js(image_dict, dict_converter=js.Object.fromEntries)
    elif isinstance(py, PointSet):
        point_set_dict = asdict(py)
        if point_set_dict["points"] is not None:
            point_set_dict["points"] = point_set_dict["points"].ravel()
        if point_set_dict["pointData"] is not None:
            point_set_dict["pointData"] = point_set_dict["pointData"].ravel()
        return pyodide.ffi.to_js(point_set_dict, dict_converter=js.Object.fromEntries)
    elif isinstance(py, Mesh):
        mesh_dict = asdict(py)
        if mesh_dict["points"] is not None:
            mesh_dict["points"] = mesh_dict["points"].ravel()
        if mesh_dict["pointData"] is not None:
            mesh_dict["pointData"] = mesh_dict["pointData"].ravel()
        if mesh_dict["cells"] is not None:
            mesh_dict["cells"] = mesh_dict["cells"].ravel()
        if mesh_dict["cellData"] is not None:
            mesh_dict["cellData"] = mesh_dict["cellData"].ravel()
        return pyodide.ffi.to_js(mesh_dict, dict_converter=js.Object.fromEntries)
    elif isinstance(py, PolyData):
        polydata_dict = asdict(py)
        if polydata_dict["points"] is not None:
            polydata_dict["points"] = polydata_dict["points"].ravel()
        if polydata_dict["vertices"] is not None:
            polydata_dict["vertices"] = polydata_dict["vertices"].ravel()
        if polydata_dict["lines"] is not None:
            polydata_dict["lines"] = polydata_dict["lines"].ravel()
        if polydata_dict["polygons"] is not None:
            polydata_dict["polygons"] = polydata_dict["polygons"].ravel()
        if polydata_dict["triangleStrips"] is not None:
            polydata_dict["triangleStrips"] = polydata_dict["triangleStrips"].ravel()
        if polydata_dict["pointData"] is not None:
            polydata_dict["pointData"] = polydata_dict["pointData"].ravel()
        if polydata_dict["cellData"] is not None:
            polydata_dict["cellData"] = polydata_dict["cellData"].ravel()
        return pyodide.ffi.to_js(polydata_dict, dict_converter=js.Object.fromEntries)
    elif isinstance(py, Transform):
        transform_dict = asdict(py)
        if transform_dict["fixedParameters"] is not None:
            transform_dict["fixedParameters"] = transform_dict["fixedParameters"].ravel()
        if transform_dict["parameters"] is not None:
            transform_dict["parameters"] = transform_dict["parameters"].ravel()
        return pyodide.ffi.to_js(transform_dict, dict_converter=js.Object.fromEntries)
    elif isinstance(py, TextStream):
        text_stream_dict = asdict(py)
        return pyodide.ffi.to_js(text_stream_dict, dict_converter=js.Object.fromEntries)
    elif isinstance(py, BinaryStream):
        binary_stream_dict = asdict(py)
        return pyodide.ffi.to_js(binary_stream_dict, dict_converter=js.Object.fromEntries)
    elif isinstance(py, BinaryFile):
        binary_file_dict = asdict(py)
        with open(py.path, "rb") as fp:
            data = fp.read()
        binary_file_dict["data"] = data
        return pyodide.ffi.to_js(binary_file_dict, dict_converter=js.Object.fromEntries)
    elif isinstance(py, TextFile):
        text_file_dict = asdict(py)
        with open(py.path, "r") as fp:
            data = fp.read()
        text_file_dict["data"] = data
        return pyodide.ffi.to_js(text_file_dict, dict_converter=js.Object.fromEntries)

    return pyodide.ffi.to_js(py, **kwargs)
